__version__ = "0.2"
__version_info__ = ['0', '2']
__version_vcs_info__ = {'branch': 'master',
 'revid': 'dca8c0f',
 'revno': 35,
 'toplevel': '/home/cstavr/okeanos/objpool'}
__version_user_info__ = "cstavr@okeanosdev-91.admin.grnet.gr"
