
__version__ = "0.11rc1"
__version_vcs_info__ = {
    'branch': 'release-0.11',
    'revid': '6778681',
    'revno': 1740}
__version_user_email__ = "saxtouri@admin.grnet.gr"
__version_user_name__ = "Stavros Sachtouris"
